/**
  *  \file ptrvector.h
  */
#ifndef PCC_V2_PTRVECTOR_H
#define PCC_V2_PTRVECTOR_H

#include <vector>
#include <utility>
#include "cpluslib/assert.h"

/** Vector of Pointers. Implements ownership semantics for a vector of
    pointers, and attempts to manage it in a type- and exception-safe way.

    The basic assumption is that ptr_vector<> owns objects passed to it.
    It will delete them when the slot holding the pointer goes away. You
    must not delete a pointer obtained from ptr_vector<>, or place it in
    another ptr_vector<>.

    Otherwise, the interface attempts to be somewhat close to
    std::vector<>, restricted to operations that are safe enough to
    maintain ownership semantics. */
template<typename T>
class ptr_vector {
    ptr_vector& operator=(const ptr_vector&);
    ptr_vector(const ptr_vector&);

    std::vector<T*> data;
 public:
    typedef typename std::vector<T*>::const_iterator iterator, const_iterator;
    typedef typename std::vector<T*>::size_type size_type;
    typedef T* value_type;

    ptr_vector();
    ~ptr_vector();

    T* push_back(T* t);
    void pop_back();
    void swap(ptr_vector& other);
    size_type size() const;
    bool empty() const;
    T* operator[](size_type index) const;
    T* front() const;
    T* back() const;

    const_iterator begin() const;
    const_iterator end() const;

    void reserve(size_type size);

    /* ptr_vector-specific functions: */
    void swap_elems(size_type a, size_type b);
    void swap_elems(size_type a, ptr_vector& other, size_type b);
    void replace_elem(size_type a, T* value);
};

/** Constructor. Makes a blank vector. */
template<typename T>
ptr_vector<T>::ptr_vector()
{ }

/** Destructor. Deletes all objects contained in the vector. */
template<typename T>
ptr_vector<T>::~ptr_vector()
{
    // delete backwards.
    for (size_type i = data.size(); i != 0; --i) {
        delete data[i-1];
        data[i-1] = 0;
    }
}

/** Add pointer at end. The pointed object becomes owned by this
    vector. If this function throws (due to lack of memory to expand
    the vector), it will delete \c t.
    \param t object to add */
template<typename T>
T*
ptr_vector<T>::push_back(T* t)
{
    try {
        data.push_back(t);
        return t;
    }
    catch (...) {
        delete t;
        throw;
    }
}

/** Remove last object. This will also delete the object.
    \pre !empty() */
template<typename T>
void
ptr_vector<T>::pop_back()
{
    ASSERT(!data.empty());
    delete data.back();
    data.back() = 0;
    data.pop_back();
}

/** Swap two ptr_vector's. */
template<typename T>
inline void
ptr_vector<T>::swap(ptr_vector<T>& other)
{
    data.swap(other.data);
}

/** Get size of this vector. */
template<typename T>
inline typename ptr_vector<T>::size_type
ptr_vector<T>::size() const
{
    return data.size();
}

/** Check whether vector is empty. */
template<typename T>
bool
ptr_vector<T>::empty() const
{
    return data.empty();
}

/** Index access.
    \param index Element index, [0,size()) */
template<typename T>
inline T*
ptr_vector<T>::operator[](size_type index) const
{
    return data[index];
}

/** Get first object. */
template<typename T>
inline T*
ptr_vector<T>::front() const
{
    return data.front();
}

/** Get last object. */
template<typename T>
inline T*
ptr_vector<T>::back() const
{
    return data.back();
}

/** Get begin iterator. */
template<typename T>
typename ptr_vector<T>::const_iterator
ptr_vector<T>::begin() const
{
    return data.begin();
}

/** Get end iterator. */
template<typename T>
typename ptr_vector<T>::const_iterator
ptr_vector<T>::end() const
{
    return data.end();
}

/** Reserve space. Make sure we can store at least \c size objects in this
    vector without re-allocation. */
template<typename T>
inline void
ptr_vector<T>::reserve(size_type size)
{
    data.reserve(size);
}

/** Swap elements.
    \param a first index
    \param b second index */
template<typename T>
inline void
ptr_vector<T>::swap_elems(size_type a, size_type b)
{
    std::swap(data[a], data[b]);
}

/** Swap elements of different ptr_vector's.
    \param a first index, pointing into this vector
    \param other other vector
    \param b other index, pointing into other vector */
template<typename T>
inline void
ptr_vector<T>::swap_elems(size_type a, ptr_vector& other, size_type b)
{
    std::swap(data[a], other.data[b]);
}

/** Replace an element. The replaced element will be deleted.
    \param a index
    \param value value to store there */
template<typename T>
void
ptr_vector<T>::replace_elem(size_type a, T* value)
{
    delete data[a];
    data[a] = value;
}

#endif
