/**
  *  \file t_fixed_vector.cc
  */

#include "cpluslib/fixed_vector.h"
#include "tests.h"
#include "cpluslib/string.h"
#include "cpluslib/format.h"

namespace{
    struct Counter {
        int n;
        static int c;

        Counter()
            : n(-9999)
            { ++c; }
        Counter(int n)
            : n(n)
            { ++c; }
        Counter(const Counter& o)
            : n(o.n)
            { ++c; }
        ~Counter()
            { --c; /*std::cout << "destroy " << n << "\n";*/ }
    };
}

int Counter::c;

static string_t
toString(const fixed_vector<Counter>& v)
{
    string_t result ="{";
    for (unsigned i = 0; i < v.size(); ++i) {
        if (i == 0)
            result += " ";
        else
            result += ", ";
        result += format("%d", v[i].n);
    }
    result += " }";
    return result;
}

void
TestFixedVector::testFixedVector()
{
    /* Creation */
    typedef Counter T;
    static_vector<T,10> is;
    TS_ASSERT_EQUALS(is.size(), 0);
    TS_ASSERT_EQUALS(Counter::c, 0);
    TS_ASSERT_EQUALS(toString(is), "{ }");

    /* push_back */
    is.push_back(1);
    is.push_back(2);
    is.push_back(3);
    is.push_back(4);
    TS_ASSERT_EQUALS(is.size(), 4);
    TS_ASSERT_EQUALS(is.front().n, 1);
    TS_ASSERT_EQUALS(is.back().n, 4);
    TS_ASSERT_EQUALS(toString(is), "{ 1, 2, 3, 4 }");
    TS_ASSERT_EQUALS(Counter::c, 4);

    /* insert */
    is.insert(is.begin() + 2, 5);
    TS_ASSERT_EQUALS(is.size(), 5);
    TS_ASSERT_EQUALS(is.front().n, 1);
    TS_ASSERT_EQUALS(is.back().n, 4);
    TS_ASSERT_EQUALS(toString(is), "{ 1, 2, 5, 3, 4 }");
    TS_ASSERT_EQUALS(Counter::c, 5);
    
    /* insert with overflow */
    TS_ASSERT_THROWS(is.insert(is.begin() + 2, static_vector<int,10>::size_type(6), 9),
                     fixed_container_overrun);
    TS_ASSERT_EQUALS(is.size(), 5);
    TS_ASSERT_EQUALS(is.front().n, 1);
    TS_ASSERT_EQUALS(is.back().n, 4);
    TS_ASSERT_EQUALS(toString(is), "{ 1, 2, 5, 3, 4 }");
    TS_ASSERT_EQUALS(Counter::c, 5);

    /* insert without overflow */
    is.insert(is.begin() + 2, static_vector<int,10>::size_type(4), 9);
    TS_ASSERT_EQUALS(is.size(), 9);
    TS_ASSERT_EQUALS(is.front().n, 1);
    TS_ASSERT_EQUALS(is.back().n, 4);
    TS_ASSERT_EQUALS(toString(is), "{ 1, 2, 9, 9, 9, 9, 5, 3, 4 }");
    TS_ASSERT_EQUALS(Counter::c, 9);

    /* erase */
    is.erase(is.begin());
    TS_ASSERT_EQUALS(is.size(), 8);
    TS_ASSERT_EQUALS(is.front().n, 2);
    TS_ASSERT_EQUALS(is.back().n, 4);
    TS_ASSERT_EQUALS(toString(is), "{ 2, 9, 9, 9, 9, 5, 3, 4 }");
    TS_ASSERT_EQUALS(Counter::c, 8);

    /* make it empty */
    static_vector<T,10>().swap(is);
    TS_ASSERT_EQUALS(is.size(), 0);
    TS_ASSERT_EQUALS(Counter::c, 0);
    TS_ASSERT_EQUALS(toString(is), "{ }");
}
